/*
 * $Id: XercesVerifier.java,v 1.3 2003/05/31 00:32:49 Bear Exp $
 *
 * Copyright 2002 Kohsuke Kawaguchi, Inc. All Rights Reserved.
 * 
 * Use is subject to license terms.
 */
package org.kohsuke.jarv.xerces;

import org.apache.xerces.impl.XMLErrorReporter;
import org.apache.xerces.impl.validation.ValidationManager;
import org.apache.xerces.impl.xs.SchemaGrammar;
import org.apache.xerces.impl.xs.XMLSchemaValidator;
import org.apache.xerces.util.NamespaceSupport;
import org.apache.xerces.util.SymbolTable;
import org.apache.xerces.util.XMLAttributesImpl;
import org.apache.xerces.xni.QName;
import org.apache.xerces.xni.XMLAttributes;
import org.apache.xerces.xni.XMLString;
import org.apache.xerces.xni.grammars.XMLGrammarPool;
import org.apache.xerces.xni.parser.XMLComponentManager;
import org.apache.xerces.xni.parser.XMLErrorHandler;
import org.apache.xerces.xni.parser.XMLParseException;
import org.iso_relax.verifier.VerifierConfigurationException;
import org.iso_relax.verifier.VerifierHandler;
import org.xml.sax.Attributes;
import org.xml.sax.ErrorHandler;
import org.xml.sax.Locator;
import org.xml.sax.SAXException;
import org.xml.sax.SAXParseException;

// TODO: allow dynamic schema loading
// TODO: handle infoset augmentation?   
/**
 * {@link Verifier} implementation for Xerces.
 * 
 * @author Kohsuke Kawaguchi (kk@kohsuke.org)
 */
public class XercesVerifier extends org.iso_relax.verifier.impl.VerifierImpl
    implements VerifierHandler
{
    /** validation will be performed against this handler. */
    private final XMLGrammarPool grammar;
    
    /** the actual validator that does all the work. */
    private final XMLSchemaValidator validator;
    
    // XMLSchemaValidator needs various helper components to work.
    private final SymbolTable symbolTable = new SymbolTable();
    private final ValidationManager validationManager = new ValidationManager();
    
    private final NamespaceSupport nsContext = new NamespaceSupport();
    
    XercesVerifier( SchemaGrammar[] _grammar ) throws VerifierConfigurationException {
        this.grammar = new ReadOnlyGrammarPool(_grammar);
        validator = new XMLSchemaValidator();
        
        validator.reset(new XMLComponentManagerImpl());
    }
    
    public VerifierHandler getVerifierHandler() {
        return this;
    }
    
    /**
     * XMLSchemaValidator needs an XMLComponentManager to work.
     */
    private class XMLComponentManagerImpl implements XMLComponentManager
    {
        XMLComponentManagerImpl() {
            errorReporter.reset(this);
        }
        
        /** error reporter is used to format error messages. */
        private final XMLErrorReporter errorReporter = new XMLErrorReporter();
        
        public Object getProperty( String propName ) {
        // TODO: ERROR_REPORTER for what?
        // TODO: XMLEntityResolver ENTITY_MANAGER for what?
            if( propName.equals(XercesConstants.SYMBOL_TABLE) )
                return symbolTable;
            if( propName.equals(XercesConstants.XMLGRAMMAR_POOL) )
                return grammar;
            if( propName.equals(XercesConstants.VALIDATION_MANAGER) )
                return validationManager;
            if( propName.equals(XercesConstants.ERROR_REPORTER) )
                return errorReporter;
            if( propName.equals(XercesConstants.ERROR_HANDLER) )
                return xercesErrorHandler;
                
            return null;    // none was found
        }
        public void setProperty( String propName, Object value ) {
            // setting properties are not allowed
            throw new UnsupportedOperationException();
        }
        public boolean getFeature( String propName ) {
            if( propName.equals(XercesConstants.SCHEMA_VALIDATION) )
                // this flag will turn on the validation
                return true;
            if( propName.equals(XercesConstants.VALIDATION) )
                return true;
            
            // TODO: how to implmenet this method?
            return false;
        }
    }
    
    private ErrorHandler getErrorHandler() {
        return super.errorHandler;
    }
    
    /**
     * XMLErroHandler adaptor
     * 
     * Receives errors through XNIErrorHandler and pass them down
     * to SAX ErrorHandler.
     * 
     * Also memorizes if there was any error.
     */
    private class XMLErrorHandlerAdaptor implements XMLErrorHandler
    {
        /** set to true if there was any error. */
        private boolean hadError = false;
        
        /**
         * returns if there was an error since the last invocation of
         * the resetError method.
         */
        public boolean hadError() { return hadError; }
        /** resets the error flag. */
        public void resetError() { hadError = false; }
        
        public void fatalError( String domain, String key, XMLParseException e ) {
            try {
                hadError = true;
                getErrorHandler().fatalError( createSAXParseException(e) );
            } catch( SAXException se ) {
                throw new WrappedSAXException(se);
            }
        }
        
        public void error( String domain, String key, XMLParseException e ) {
            try {
                hadError = true;
                getErrorHandler().error( createSAXParseException(e) );
            } catch( SAXException se ) {
                throw new WrappedSAXException(se);
            }
        }
        
        public void warning( String domain, String key, XMLParseException e ) {
            try {
                getErrorHandler().warning( createSAXParseException(e) );
            } catch( SAXException se ) {
                throw new WrappedSAXException(se);
            }
        }
        
        private SAXParseException createSAXParseException( XMLParseException e ) {
            return new SAXParseException( e.getMessage(),
                e.getExpandedSystemId(), e.getPublicId(),
                e.getLineNumber(), e.getColumnNumber() );
        }
    }
    
    private final XMLErrorHandlerAdaptor xercesErrorHandler
        = new XMLErrorHandlerAdaptor();
    
    
    /** Symbolizes the specified string. */
    private String symbolize( String s ) {
        if(s==null)     return null;
        else            return symbolTable.addSymbol(s);
    }
    
    /** Creates a QName object. */
    private QName createQName( String uri, String local, String raw ) {
        int idx = raw.indexOf(':');
        String prefix;
        if(idx<0) prefix=null;
        else      prefix=raw.substring(0,idx);
        
        if(uri.length()==0)
            uri=null; // XNI uses null whereas SAX uses the empty string
        
        return new QName(
            symbolize(prefix), symbolize(local),
            symbolize(raw), symbolize(uri) );  
    }
    
    /** only one instance of XMLAttributes is used. */
    private final XMLAttributes xa = new XMLAttributesImpl();
    
    /** Creates an XMLAttributes object. */
    private XMLAttributes createAttributes( Attributes att ) {
        // if the parser is Xerces, just use the Attributes as is.
        if(att instanceof XMLAttributes)
            return (XMLAttributes)att;
        
        // otherwise use our local copy
        xa.removeAllAttributes();
        int len = att.getLength();
        for( int i=0; i<len; i++ )
            xa.addAttribute(createQName(
                att.getURI(i),att.getLocalName(i),att.getQName(i)),
                att.getType(i),att.getValue(i));
        return xa;
    }
    
    private XMLString createXMLString( String str ) {
        // with my patch
        // return new XMLString(str);
        
        // for now
        return new XMLString(str.toCharArray(),0,str.length());
    }   
    
    public boolean isValid() {
        return !xercesErrorHandler.hadError();
    }
    
    
    //
    // ContentHandler implementation
    //
    public void startDocument() {
        // reset the error flag when we start a new validation.
        xercesErrorHandler.resetError();
        nsContext.reset();
        
        // TODO: what parameters do I need to pass?
        validator.startDocument(
            (locator==null)?null:new LocatorAdaptor(locator),
            null,
            nsContext,
            null);
    }
    
    public void endDocument() {
        validator.endDocument(null);
    }
    
    public void startElement( String uri, String local, String qname, Attributes att ) {
        validator.startElement(createQName(uri,local,qname),createAttributes(att),null);
    }
    
    public void endElement( String uri, String local, String qname ) {
        validator.endElement(createQName(uri,local,qname),null);
    }
    
    public void characters( char[] buf, int offset, int len ) {
        validator.characters(new XMLString(buf,offset,len),null);
    }
    
    public void ignorableWhitespace( char[] buf, int offset, int len ) {
        validator.ignorableWhitespace(new XMLString(buf,offset,len),null);
    }
    
    public void startPrefixMapping( String prefix, String uri ) {
        nsContext.pushContext();
        nsContext.declarePrefix(prefix,uri);
    }
    
    public void endPrefixMapping( String prefix ) {
        nsContext.popContext();
    }
    
    public void processingInstruction( String target, String data ) {
        validator.processingInstruction(
            symbolize(target),createXMLString(data),null);
    }
    
    public void skippedEntity( String name ) {
        // TODO: what method do I need to call?
        // validator.startGeneralEntity?
    }
    
    private Locator locator;
    public void setDocumentLocator( Locator _loc ) {
        this.locator = _loc;
    }
}
